
SV * globalMinDifferences_helper(char *s1, char *s2);
SV * boundedGlobalMinDifferences_helper(char *s1, char *s2, int bound);
SV * boundedHirschbergGlobalMinDiffs_helper(char *s1, char *s2, int bound);

