#ifndef XTEND1_H
#define XTEND1_H

/* $Id: Xtend1.h,v 1.1 2002/12/03 20:12:37 hartzell Exp $ */
extern int    Xextend_bw(uchar *,uchar *,int,int,int,int,int *,int *);
extern int    Xextend_fw(uchar *,uchar *,int,int,int,int,int *,int *);
#endif /* XTEND1_H */
