#ifndef LIBC_H
#define LIBC_H
/*$Id: libc.h,v 1.1 2002/12/03 20:12:37 hartzell Exp $*/
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif
