# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl test.pl'

######################### We start with some black magic to print on failure.

# Change 1..1 below to 1..last_test_to_print .
# (It may become useful if the test is moved to ./t subdirectory.)

BEGIN { $| = 1; print "1..14\n"; }
END {print "not ok 1\n" unless $loaded;}
use GH::EditOp;
$loaded = 1;
print "ok 1\n";

######################### End of black magic.

# Insert your test code below (better if it prints "ok 13"
# (correspondingly "not ok 13") depending on the success of chunk 13
# of the test code):

sub Not {
  print "not ";
}

sub Ok {
  my($i) = @_;
  print "ok $i\n";
}
  
$i = 2;

Not() if (NOP != 0); Ok($i++);

my $editop = new GH::EditOp;

$editop->setType(NOP);
Not() if ($editop->getType() != NOP); Ok($i++);
Not() if ($editop->getOpName() ne "nop"); Ok($i++);

$editop->setType(MATCH);
Not() if ($editop->getType() != MATCH); Ok($i++);
Not() if ($editop->getOpName() ne "match"); Ok($i++);

$editop->setType(MISMATCH);
Not() if ($editop->getType() != MISMATCH); Ok($i++);
Not() if ($editop->getOpName() ne "mismatch"); Ok($i++);

$editop->setType(INSERT_S1);
Not() if ($editop->getType() != INSERT_S1); Ok($i++);
Not() if ($editop->getOpName() ne "insert_s1"); Ok($i++);

$editop->setType(INSERT_S2);
Not() if ($editop->getType() != INSERT_S2); Ok($i++);
Not() if ($editop->getOpName() ne "insert_s2"); Ok($i++);

$editop->setCount(0);
Not() if ($editop->getCount() != 0); Ok($i++);

$editop->setCount(2345);
Not() if ($editop->getCount() != 2345); Ok($i++);


